@@ -16,6 +16,7 @@ import logging
 import os
 import select
 import socket
+import ssl
 import struct
 import threading
 import time
@@ -28,6 +29,11 @@ try:
 except ImportError:
     import queue as queue
 
+try:
+    import statsd
+except ImportError:
+    statsd = None
+
 PRECEDENCE_NORMAL = 0
 PRECEDENCE_LOW = 1
 PRECEDENCE_HIGH = 2
@@ -106,9 +112,16 @@ class Connection(object):
 
     log = logging.getLogger("gear.Connection")
 
-    def __init__(self, host, port):
+    def __init__(self, host, port, ssl_key=None, ssl_cert=None, ssl_ca=None):
         self.host = host
         self.port = port
+        self.ssl_key = ssl_key
+        self.ssl_cert = ssl_cert
+        self.ssl_ca = ssl_ca
+
+        self.use_ssl = False
+        if all([self.ssl_key, self.ssl_cert, self.ssl_ca]):
+            self.use_ssl = True
 
         self.echo_lock = threading.Lock()
         self._init()
@@ -153,6 +166,15 @@ class Connection(object):
             except socket.error:
                 s = None
                 continue
+
+            if self.use_ssl:
+                self.log.debug("Using SSL")
+                s = ssl.wrap_socket(s, ssl_version=ssl.PROTOCOL_TLSv1,
+                                    cert_reqs=ssl.CERT_REQUIRED,
+                                    keyfile=self.ssl_key,
+                                    certfile=self.ssl_cert,
+                                    ca_certs=self.ssl_ca)
+
             try:
                 s.connect(sa)
             except socket.error:
@@ -190,17 +212,57 @@ class Connection(object):
         self.disconnect()
         self.connect()
 
+    def sendRaw(self, data):
+        """Send raw data over the socket.
+
+        :arg bytes data The raw data to send
+        """
+        while True:
+            try:
+                self.conn.send(data)
+            except ssl.SSLError as e:
+                if e.errno == ssl.SSL_ERROR_WANT_READ:
+                    continue
+                elif e.errno == ssl.SSL_ERROR_WANT_WRITE:
+                    continue
+                else:
+                    raise
+            break
+
     def sendPacket(self, packet):
         """Send a packet to the server.
 
         :arg Packet packet: The :py:class:`Packet` to send.
         """
         self.log.debug("Sending packet: %s" % packet)
-        self.conn.send(packet.toBinary())
+        self.sendRaw(packet.toBinary())
 
     def _getAdminRequest(self):
         return self.admin_requests.pop(0)
 
+    def _readRawBytes(self, bytes_to_read):
+        while True:
+            try:
+                buff = self.conn.recv(bytes_to_read)
+            except ssl.SSLError as e:
+                if e.errno == ssl.SSL_ERROR_WANT_READ:
+                    continue
+                elif e.errno == ssl.SSL_ERROR_WANT_WRITE:
+                    continue
+                else:
+                    raise
+            break
+
+        bytes_read = len(buff)
+        if self.use_ssl and (bytes_read < bytes_to_read):
+            remaining = self.conn.pending()
+            while remaining and (bytes_read < bytes_to_read):
+                buff += self.conn.recv(bytes_to_read - bytes_read)
+                remaining = self.conn.pending()
+                bytes_read = len(buff)
+
+        return buff
+
     def readPacket(self):
         """Read one packet or administrative response from the server.
 
@@ -216,7 +278,7 @@ class Connection(object):
         admin = None
         admin_request = None
         while True:
-            c = self.conn.recv(1)
+            c = self._readRawBytes(1)
             if not c:
                 return None
             if admin is None:
@@ -245,7 +307,7 @@ class Connection(object):
             is received.
         """
         self.admin_requests.append(request)
-        self.conn.send(request.getCommand())
+        self.sendRaw(request.getCommand())
         complete = request.waitForResponse(timeout)
         if not complete:
             raise TimeoutError()
@@ -926,7 +988,8 @@ class BaseClient(BaseClientServer):
         # connection state, such as setting options or functions.
         self.broadcast_lock = threading.RLock()
 
-    def addServer(self, host, port=4730):
+    def addServer(self, host, port=4730,
+                  ssl_key=None, ssl_cert=None, ssl_ca=None):
         """Add a server to the client's connection pool.
 
         Any number of Gearman servers may be added to a client.  The
@@ -941,8 +1004,13 @@ class BaseClient(BaseClientServer):
         ensure that a connection is ready before proceeding, see
         :py:meth:`waitForServer`.
 
+        When using SSL connections, all SSL files must be specified.
+
         :arg str host: The hostname or IP address of the server.
         :arg int port: The port on which the gearman server is listening.
+        :arg str ssl_key: Path to the SSL private key.
+        :arg str ssl_cert: Path to the SSL certificate.
+        :arg str ssl_ca: Path to the CA certificate.
         :raises ConfigurationError: If the host/port combination has
             already been added to the client.
         """
@@ -954,7 +1022,7 @@ class BaseClient(BaseClientServer):
             for conn in self.active_connections + self.inactive_connections:
                 if conn.host == host and conn.port == port:
                     raise ConfigurationError("Host/port already specified")
-            conn = Connection(host, port)
+            conn = Connection(host, port, ssl_key, ssl_cert, ssl_ca)
             self.inactive_connections.append(conn)
             self.connections_condition.notifyAll()
         finally:
@@ -1373,8 +1441,8 @@ class Client(BaseClient):
         """
 
         job = packet.getJob()
-        job.known = (packet.getArgument(1) == 1)
-        job.running = (packet.getArgument(2) == 1)
+        job.known = (packet.getArgument(1) == '1')
+        job.running = (packet.getArgument(2) == '1')
         job.numerator = packet.getArgument(3)
         job.denominator = packet.getArgument(4)
 
@@ -2019,11 +2087,11 @@ class ServerAdminRequest(AdminRequest):
 class ServerConnection(Connection):
     """A Connection to a Gearman Client."""
 
-    def __init__(self, addr, conn):
+    def __init__(self, addr, conn, use_ssl):
         self.host = addr[0]
         self.port = addr[1]
         self.conn = conn
-        self.max_handle = 0
+        self.use_ssl = use_ssl
         self.client_id = None
         self.functions = set()
         self.related_jobs = {}
@@ -2046,17 +2114,43 @@ class Server(BaseClientServer):
     (not for production use).
 
     :arg int port: The TCP port on which to listen.
+    :arg str ssl_key: Path to the SSL private key.
+    :arg str ssl_cert: Path to the SSL certificate.
+    :arg str ssl_ca: Path to the CA certificate.
+    :arg str statsd_host: statsd hostname.  None means disabled
+        (the default).
+    :arg str statsd_port: statsd port (defaults to 8125).
+    :arg str statsd_prefix: statsd key prefix.
     """
 
-    def __init__(self, port=4730):
+    def __init__(self, port=4730, ssl_key=None, ssl_cert=None, ssl_ca=None,
+                 statsd_host=None, statsd_port=8125, statsd_prefix=None):
         self.port = port
+        self.ssl_key = ssl_key
+        self.ssl_cert = ssl_cert
+        self.ssl_ca = ssl_ca
+        if statsd_host:
+            if not statsd:
+                self.log.error("Unable to import statsd module")
+                self.statsd = None
+            else:
+                self.statsd = statsd.StatsClient(statsd_host,
+                                                 statsd_port,
+                                                 statsd_prefix)
+        else:
+            self.statsd = None
         self.high_queue = []
         self.normal_queue = []
         self.low_queue = []
         self.jobs = {}
         self.functions = set()
+        self.max_handle = 0
         self.connect_wake_read, self.connect_wake_write = os.pipe()
 
+        self.use_ssl = False
+        if all([self.ssl_key, self.ssl_cert, self.ssl_ca]):
+            self.use_ssl = True
+
         for res in socket.getaddrinfo(None, self.port, socket.AF_UNSPEC,
                                       socket.SOCK_STREAM, 0,
                                       socket.AI_PASSIVE):
@@ -2113,7 +2207,14 @@ class Server(BaseClientServer):
                     self.log.debug("Accepting new connection")
                     c, addr = self.socket.accept()
                     self.log.debug("Accepted new connection")
-                    conn = ServerConnection(addr, c)
+                    if self.use_ssl:
+                        c = ssl.wrap_socket(c, server_side=True,
+                                            keyfile=self.ssl_key,
+                                            certfile=self.ssl_cert,
+                                            ca_certs=self.ssl_ca,
+                                            cert_reqs=ssl.CERT_REQUIRED,
+                                            ssl_version=ssl.PROTOCOL_TLSv1)
+                    conn = ServerConnection(addr, c, self.use_ssl)
                     self.connections_condition.acquire()
                     self.active_connections.append(conn)
                     self.connections_condition.notifyAll()
@@ -2149,6 +2250,7 @@ class Server(BaseClientServer):
             if job.worker_connection:
                 del job.worker_connection.related_jobs[job.handle]
             del self.jobs[job.handle]
+        self._updateStats()
 
     def getQueue(self):
         """Returns a copy of all internal queues in a flattened form.
@@ -2179,11 +2281,12 @@ class Server(BaseClientServer):
                     if handle == job.handle:
                         queue.remove(job)
                         del self.jobs[handle]
-                        request.connection.conn.send(b'OK\n')
+                        self._updateStats()
+                        request.connection.sendRaw(b'OK\n')
                         return
-        request.connection.conn.send(b'ERR UNKNOWN_JOB\n')
+        request.connection.sendRaw(b'ERR UNKNOWN_JOB\n')
 
-    def handleStatus(self, request):
+    def _getFunctionStats(self):
         functions = {}
         for function in self.functions:
             # Total, running, workers
@@ -2195,11 +2298,15 @@ class Server(BaseClientServer):
         for connection in self.active_connections:
             for function in connection.functions:
                 functions[function][2] += 1
+        return functions
+
+    def handleStatus(self, request):
+        functions = self._getFunctionStats()
         for name, values in functions.items():
-            request.connection.conn.send(("%s\t%s\t%s\t%s\n" %
-                                         (name, values[0], values[1],
-                                          values[2])).encode('utf8'))
-        request.connection.conn.send(b'.\n')
+            request.connection.sendRaw(("%s\t%s\t%s\t%s\n" %
+                                       (name, values[0], values[1],
+                                        values[2])).encode('utf8'))
+        request.connection.sendRaw(b'.\n')
 
     def handleWorkers(self, request):
         for connection in self.active_connections:
@@ -2207,17 +2314,58 @@ class Server(BaseClientServer):
             ip = connection.host
             client_id = connection.client_id or '-'
             functions = ' '.join(connection.functions)
-            request.connection.conn.send(("%s %s %s : %s\n" %
-                                          (fd, ip, client_id, functions))
-                                         .encode('utf8'))
-        request.connection.conn.send(b'.\n')
+            request.connection.sendRaw(("%s %s %s : %s\n" %
+                                       (fd, ip, client_id, functions))
+                                       .encode('utf8'))
+        request.connection.sendRaw(b'.\n')
 
     def wakeConnections(self):
         p = Packet(constants.RES, constants.NOOP, b'')
         for connection in self.active_connections:
             if connection.state == 'SLEEP':
-                connection.sendPacket(p)
                 connection.changeState("AWAKE")
+                connection.sendPacket(p)
+
+    def _updateStats(self):
+        if not self.statsd:
+            return
+
+        # prefix.queue.JOB.waiting
+        # prefix.queue.JOB.running
+        # prefix.queue.JOB.workers
+        # prefix.queue.waiting
+        # prefix.queue.running
+        # prefix.queue.workers
+        functions = self._getFunctionStats()
+        base_key = 'queue'
+        total_waiting = 0
+        total_running = 0
+        for name, values in functions.items():
+            (total, running, workers) = values
+            job_key = '.'.join([base_key, name])
+
+            key = '.'.join([job_key, 'waiting'])
+            self.statsd.gauge(key, total - running)
+            total_waiting += (total - running)
+
+            key = '.'.join([job_key, 'running'])
+            self.statsd.gauge(key, running)
+            total_running += running
+
+            key = '.'.join([job_key, 'workers'])
+            self.statsd.gauge(key, workers)
+
+        key = '.'.join([base_key, 'waiting'])
+        self.statsd.gauge(key, total_waiting)
+        key = '.'.join([base_key, 'running'])
+        self.statsd.gauge(key, total_running)
+
+        total_workers = 0
+        for connection in self.active_connections:
+            if connection.functions:
+                total_workers += 1
+        key = '.'.join([base_key, 'workers'])
+        self.statsd.gauge(key, total_workers)
 
     def _handleSubmitJob(self, packet, precedence):
         name = packet.getArgument(0)
@@ -2225,9 +2373,9 @@ class Server(BaseClientServer):
         if not unique:
             unique = None
         arguments = packet.getArgument(2, True)
-        packet.connection.max_handle += 1
+        self.max_handle += 1
         handle = ('H:%s:%s' % (packet.connection.host,
-                               packet.connection.max_handle)).encode('utf8')
+                               self.max_handle)).encode('utf8')
         job = ServerJob(handle, name, arguments, packet.connection, unique)
         p = Packet(constants.RES, constants.JOB_CREATED, handle)
         packet.connection.sendPacket(p)
@@ -2239,6 +2387,7 @@ class Server(BaseClientServer):
             self.normal_queue.append(job)
         elif precedence == PRECEDENCE_LOW:
             self.low_queue.append(job)
+        self._updateStats()
         self.wakeConnections()
 
     def handleSubmitJob(self, packet):
@@ -2259,6 +2408,7 @@ class Server(BaseClientServer):
                         connection.related_jobs[job.handle] = job
                         job.worker_connection = connection
                     job.running = True
+                    self._updateStats()
                     return job
         return None
 
@@ -2321,6 +2471,7 @@ class Server(BaseClientServer):
             del self.jobs[handle]
             del job.client_connection.related_jobs[handle]
             del job.worker_connection.related_jobs[handle]
+            self._updateStats()
 
     def handleSetClientID(self, packet):
         name = packet.getArgument(0)
